@@ -152,7 +152,8 @@ tds_process_default_tokens(TDSSOCKET * tds, uint8_t marker)
 	case TDS_RETURNSTATUS_TOKEN:
 		ret_status = tds_get_int(tds);
 		marker = tds_peek(tds);
-		if (marker != TDS_PARAM_TOKEN && marker != TDS_DONEPROC_TOKEN && marker != TDS_DONE_TOKEN)
+		if (marker != TDS_PARAM_TOKEN && marker != TDS_DONEPROC_TOKEN && marker != TDS_DONE_TOKEN
+		    && marker != TDS5_PARAMFMT_TOKEN)
 			break;
 		tds->has_status = true;
 		tds->ret_status = ret_status;
@@ -1159,9 +1160,8 @@ tds71_read_table_names(TDSSOCKET *tds, int remainder, struct namelist **p_head)
 		/* allocate full name */
 		p = tds_new(char, len);
 		if (!p) {
-			i = elements;
-			while (i > 0)
-				free(partials[--i]);
+			for (i = 0; i < elements; ++i)
+				free(partials[i]);
 			tds_free_namelist(head);
 			return -1;
 		}
